# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stochastic rounding utilities."""

from typing import Sequence
import jax


def uniform_noise(
    shape: tuple[int, ...],
    *,
    key: jax.Array,
    channelwise_noise_axes: Sequence[int] = (0,),
) -> jax.Array:
  """Uniform noise."""

  # Keep shape dimensions only for channelwise_noise_axes.
  noise_shape = tuple(
      dim if axis in channelwise_noise_axes else 1
      for axis, dim in enumerate(shape)
  )
  return jax.random.uniform(key, noise_shape) - 0.5



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantized jax.lax.ragged_dot."""

import jax
from jax import numpy as jnp
from qwix._src.core import qarray


def _fast_ragged_dot(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
) -> jax.Array:
  """Quantized jax.lax.ragged_dot."""
  if isinstance(lhs, qarray.QArray):
    assert lhs.zero_point is None, 'not supported yet'
    lhs_value = lhs.qvalue
    lhs_scale = lhs.scale
  else:
    lhs_value = lhs
    lhs_scale = None
  if isinstance(rhs, qarray.QArray):
    assert rhs.zero_point is None, 'not supported yet'
    rhs_value = rhs.qvalue
    rhs_scale = rhs.scale
  else:
    rhs_value = rhs
    rhs_scale = None

  preferred_element_type, result_type = qarray.get_accumulator_and_result_type(
      lhs, rhs, preferred_element_type=preferred_element_type
  )

  out = jax.lax.ragged_dot(
      lhs_value,
      rhs_value,
      group_sizes,
      precision=precision,
      preferred_element_type=preferred_element_type,
      group_offset=group_offset,
  )

  # ragged_dot has fixed dimension numbers which makes implementation a lot
  # easier, i.e., lhs: [m, k], rhs: [g, k, n], res: [m, n].
  # TODO(dangyi): support arbitrary dimension numbers.
  if lhs_scale is not None:  # [m, 1]
    lhs_scale = qarray.transpose_array(lhs_scale, (0, None))
    out = qarray.call_with_generic_broadcast(jnp.multiply, out, lhs_scale)
  if rhs_scale is not None:  # [1, 1, n] or [g, 1, n]
    if rhs_scale.shape[0] == 1:
      # It's possible to apply the scale to the out directly.
      rhs_scale = qarray.transpose_array(rhs_scale, (None, 2))
    else:
      # We need another ragged_dot to apply the scale to the out.
      rhs_scale = jax.lax.ragged_dot(
          jnp.ones((out.shape[0], 1), rhs_scale.dtype),
          rhs_scale,
          group_sizes,
          group_offset=group_offset,
      )
    out = qarray.call_with_generic_broadcast(jnp.multiply, out, rhs_scale)

  return out.astype(result_type)


def _slow_ragged_dot(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    **kwargs,
) -> jax.Array:
  """Quantized jax.lax.ragged_dot which dequantizes first."""
  if isinstance(lhs, qarray.QArray):
    lhs = qarray.dequantize(lhs)
  if isinstance(rhs, qarray.QArray):
    rhs = qarray.dequantize(rhs)
  return jax.lax.ragged_dot(lhs, rhs, group_sizes, **kwargs)


def ragged_dot(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
) -> jax.Array:
  """Quantized jax.lax.ragged_dot."""
  use_fast_ragged_dot = True

  # fast_ragged_dot does't support channelwise scales on group axis, or tiled
  # scales on contracting axes, or zero_point.
  if isinstance(lhs, qarray.QArray):  # [m, k]
    if lhs.zero_point is not None or lhs.scale.shape[1] > 1:
      use_fast_ragged_dot = False
  if isinstance(rhs, qarray.QArray):  # [g, k, n]
    if (
        rhs.zero_point is not None
        or rhs.scale.shape[0] > 1
        or rhs.scale.shape[1] > 1
    ):
      use_fast_ragged_dot = False

  if use_fast_ragged_dot:
    return _fast_ragged_dot(
        lhs,
        rhs,
        group_sizes,
        precision=precision,
        preferred_element_type=preferred_element_type,
        group_offset=group_offset,
    )
  else:
    return _slow_ragged_dot(
        lhs,
        rhs,
        group_sizes,
        precision=precision,
        preferred_element_type=preferred_element_type,
        group_offset=group_offset,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qwix APIs for quantizing Pallas kernels.

This module is usually imported as `import qwix.pallas as qpl` in user code.

To add quantization support to a Pallas kernel, generally users need to

1) Quantize the inputs outside of the kernel via `qpl.quantize`.
2) Instead of calling `pl.pallas_call`, call `qpl.pallas_call`, which supports
   taking QArrays as inputs. `qpl.pallas_call` takes the same block specs as
   `pl.pallas_call` and adjusts them automatically for QArrays.
3) When inside the kernel, use `qpl.einsum` / `qpl.dot_general` / `qpl.dot`
   which take QArrays as inputs and return jax.Array. Alternatively, call
   `dequantize` on the QArray to obtain the dequantized jax.Array.

Note: dot_general / einsum / dot in this module should only be called inside
pallas kernels.
"""

__all__ = [
    'quantize',
    'dot',
    'dot_general',
    'einsum',
    'pallas_call',
    'transform_block_specs_for_tpu',
    'update_block_specs_for_qarray',
    'QArray',
    'dequantize',
    'get_current_rule',
]

# pylint: disable=g-multiple-import, g-importing-member

from collections.abc import Collection, Mapping
import dataclasses
import functools

import jax
from qwix._src.core import qarray
from qwix._src.core.dot_general import loop_dot_general as dot_general
from qwix._src.core.einsum import einsum as _einsum
from qwix._src.core.pallas import pallas_call, transform_block_specs_for_tpu, update_block_specs_for_qarray
from qwix._src.core.qarray import QArray, dequantize
from qwix._src.qconfig import get_current_rule

einsum = functools.partial(_einsum, _qwix_dot_general=dot_general)


def quantize(
    array: jax.Array,
    qtype: jax.typing.DTypeLike,
    *,
    channelwise_axes: Collection[int] = (),
    tiled_axes: Mapping[int, int | float] | None = None,
    calibration_method: str = 'absmax',
    scale_dtype: jax.typing.DTypeLike | None = None,
) -> QArray:
  """Quantize a Jax Array into QArray.

  Args:
    array: The array to quantize.
    qtype: The logical type of the quantized value, e.g. jnp.int8, jnp.int4,
      jnp.float8_e4m3fn, "nf4", etc.
    channelwise_axes: Channelwise axes have individual scales. This has the same
      effect as setting their tile sizes to 1 in tiled_axes.
    tiled_axes: Tiled axes have blockwise scales, aka subchannel quantization.
      The value is a mapping from the tiled axis to the tile size. If the tile
      size is a float, it will be interpreted as "1 / tile_count" and the actual
      tile size will be round(axis_size * tile_size).
    calibration_method: The calibration method to use. The format is
      "<method>[,<args>]", e.g. "absmax" or "fixed,-10,10".
    scale_dtype: The dtype of the scale. If not given, the dtype will be the
      same as the array's dtype.

  Returns:
    The quantized array.
  """
  # A stable API for qarray.quantize()
  how = qarray.HowToQuantize(
      qtype=qtype,
      channelwise_axes=channelwise_axes,
      tiled_axes=tiled_axes or {},
      calibration_method=calibration_method,
  )
  array = qarray.quantize(array, how)
  if scale_dtype is not None:
    array = dataclasses.replace(array, scale=array.scale.astype(scale_dtype))
  return array


def dot(
    a: jax.Array | QArray,
    b: jax.Array | QArray,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    out_sharding=None,
):
  """jnp.dot with QArray support."""
  a_ndim = a.ndim
  b_ndim = b.ndim
  if a_ndim == 0 or b_ndim == 0:
    contract_dims = ((), ())
  elif b_ndim == 1:
    contract_dims = ((a_ndim - 1,), (0,))
  else:
    contract_dims = ((a_ndim - 1,), (b_ndim - 2,))
  return dot_general(
      a,
      b,
      dimension_numbers=(contract_dims, ((), ())),
      precision=precision,
      preferred_element_type=preferred_element_type,
      out_sharding=out_sharding,
  )
